use serde::{Deserialize, Serialize};
use serde_json;

use crate::subscription;
use crate::database;
use crate::error::*;
use crypto::digest::Digest;

pub struct Envelope {
    pub endpoint: String,
    pub from_domain: String,
    pub to_domain: String,
    pub authorization: Option<String>,
    pub data: serde_json::Value,
}

pub fn endpoint(envelope: Envelope) -> Result<serde_json::Value, Error> {
    // TODO check authorization
    // TODO domain_check
    match envelope.endpoint.as_str() {
        "echo" => Ok(serde_json::to_value(echo(serde_json::from_value(envelope.data)?)?)?),
        "authenticate" => Ok(serde_json::to_value(authenticate(serde_json::from_value(envelope.data)?)?)?),
        "subscribe" => Ok(serde_json::to_value(subscribe(serde_json::from_value(envelope.data)?)?)?),
        "send_event" => Ok(serde_json::to_value(send_event(serde_json::from_value(envelope.data)?)?)?),
        _ => return Err(Error::new(Kind::InvalidEndpointError, Class::ClientError)),
    }
}

pub fn get_id(input: &str) -> String {
    let mut hasher = crypto::sha2::Sha256::new();
    hasher.input_str(chrono::Utc::now().timestamp_millis().to_string().as_str());
    hasher.input_str(" ");
    hasher.input_str(input);

    let mut result = [0u8; 32];
    hasher.result(&mut result);
    base64_url::encode(&result)
}

#[derive(Serialize, Deserialize)]
pub struct EchoInput {
    message: String,
}

#[derive(Serialize, Deserialize)]
pub struct EchoOutput {
    message: String,
    database: Option<i32>,
}

pub fn echo(input: EchoInput) -> Result<EchoOutput, Error> {
    let backend = database::client()?;
    let mut output = EchoOutput {
        message: input.message,
        database: None,
    };
    match backend {
        database::Client::Postgres(mut client) => {
            let res = client.query("SELECT * FROM test", &[])?;
            for row in res {
                output.database = Some(row.get(0));
            }
        }
    }
    Ok(output)
}

#[derive(Serialize, Deserialize)]
pub struct AuthenticateInput {
    r#type: String,
    name: String,
    password: String,
}

#[derive(Serialize, Deserialize)]
pub struct AuthenticateOutput {
    token: String,
}

pub fn authenticate(input: AuthenticateInput) -> Result<AuthenticateOutput, Error> {
    let backend = database::client()?;

    let mut token;
    match backend {
        database::Client::Postgres(mut client) => {
            let res = client.query("SELECT account_id FROM accounts WHERE account_name = $1", &[&input.name])?;
            if res.len() == 0 {
                return Err(Error::new(Kind::AuthenticationError, Class::ClientError));
            }
            token = res.get(0).unwrap().get(0);
        }
    }

    Ok(AuthenticateOutput { token })
}

#[derive(Serialize, Deserialize)]
pub struct SendEventInput {
    room_id: String,
    data: serde_json::Value,
}

#[derive(Serialize, Deserialize)]
pub struct SendEventOutput {
    event_id: String,
}

pub fn send_event(input: SendEventInput) -> Result<SendEventOutput, Error> {
    let backend = database::client()?;
    let event_id = get_id("hermann");  // TODO fix id generation
    let data = serde_json::to_string(&input.data)?;

    match backend {
        database::Client::Postgres(mut client) => {
            client.execute("INSERT INTO events (event_id, room_id, data) VALUES ($1, $2, to_jsonb($3::text))", &[&event_id, &input.room_id, &data])?;
        }
    }

    subscription::notify(subscription::Event {
        data: input.data
    });

    Ok(SendEventOutput { event_id })
}

#[derive(Serialize, Deserialize)]
pub struct SubscribeInput {

}

#[derive(Serialize, Deserialize)]
pub struct SubscribeOutput {
    event: subscription::Event,
}

pub fn subscribe(input: SubscribeInput) -> Result<SubscribeOutput, Error> {
    let rx = subscription::subscribe();
    let event = rx.recv().unwrap();
    subscription::unsubscribe(rx);
    Ok(SubscribeOutput { event })
}
