use crate::error::*;
use crate::http;

use crate::usimp;
use crate::websocket::*;
use base64;
use crypto;
use crypto::digest::Digest;

pub fn recv_message(client: &mut http::HttpStream) -> Result<Message, Error> {
    let mut msg: Vec<u8> = Vec::new();
    let mut msg_type = 0;
    loop {
        let header = FrameHeader::from(&mut client.stream)?;

        // FIXME control frames may show up in a fragmented stream
        if msg_type != 0 && header.opcode != 0 {
            return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("continuation frame expected".to_string()));
        } else if header.opcode >= 8 && (!header.fin || header.payload_len >= 126) {
            return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("invalid control frame".to_string()));
        }

        match header.opcode {
            0 => {},  // cont
            1 => {},  // text
            2 =>   // binary
                return Err(Error::new(Kind::UsimpProtocolError, Class::ClientError)
                    .set_desc("binary frames must not be sent on a usimp connection".to_string())),
            8 => {},  // close
            9 => {},  // ping
            10 => {},  // pong
            _ => return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("invalid opcode".to_string())),
        }

        msg_type = header.opcode;

        // FIXME check payload len and total len

        let mut buf = vec![0u8; header.payload_len() as usize];
        client.stream.read_exact(&mut buf)?;

        if header.mask {
            let key: [u8; 4] = [
                (header.masking_key.unwrap() >> 24) as u8,
                ((header.masking_key.unwrap() >> 16) & 0xFF) as u8,
                ((header.masking_key.unwrap() >> 8) & 0xFF) as u8,
                (header.masking_key.unwrap() & 0xFF) as u8,
            ];
            for (pos, byte) in buf.iter_mut().enumerate() {
                *byte ^= key[pos & 3];  // = pos % 4
            }
        }

        msg.append(&mut buf);

        if header.fin {
            break
        }
    }

    match msg_type {
        1 => {Ok(Message::TextMessage(TextMessage {
            data: String::from_utf8(msg)?
        }))},
        8 => {
            let mut code = None;
            let mut reason = None;

            if msg.len() >= 2 {
                code = Some(((msg[0] as u16) << 8) | (msg[1] as u16));
            }

            if msg.len() > 2 {
                reason = Some(String::from_utf8(msg[2..].to_vec())?);
            }

            Ok(Message::CloseMessage(CloseMessage {
                code,
                reason
            }))
        },
        9 => {Ok(Message::PingMessage(PingMessage {
            data: String::from_utf8(msg)?
        }))},
        10 => {Ok(Message::PongMessage(PongMessage {
            data: String::from_utf8(msg)?
        }))},
        _ => panic!("invalid msg_type for websocket")
    }
}

pub fn handshake(
    client: &mut http::HttpStream,
    req: &http::Request,
    res: &mut http::Response,
) -> Result<(), Error> {
    if let http::Method::GET = req.method {
    } else {
        res.status(405);
        res.header.add_field("Allow", "GET");
        return Err(Error::new(Kind::WebSocketError, Class::ClientError)
            .set_desc("method not allowed".to_string()));
    }

    if let Some(_) = req.header.find_field("Connection") {
        if !req.header.field_has_value("Connection", "upgrade") {
            return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("invalid value for header field 'Connection'".to_string()));
        }
    } else {
        return Err(Error::new(Kind::WebSocketError, Class::ClientError)
            .set_desc("unable to find header field 'Connection'".to_string()));
    }

    if let Some(upgrade) = req.header.find_field("Upgrade") {
        if !upgrade.eq_ignore_ascii_case("websocket") {
            return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("invalid value for header field 'Upgrade'".to_string()));
        }
    } else {
        return Err(Error::new(Kind::WebSocketError, Class::ClientError)
           .set_desc("unable to find header field 'Upgrade'".to_string()));
    }

    if let Some(version) = req.header.find_field("Sec-WebSocket-Version") {
        if !version.eq("13") {
            return Err(Error::new(Kind::WebSocketError, Class::ClientError)
                .set_desc("invalid value for header field 'Sec-WebSocket-Key'".to_string()));
        }
    } else {
        return Err(Error::new(Kind::WebSocketError, Class::ClientError)
            .set_desc("unable to find header field 'Sec-WebSocket-Version'".to_string()));
    }

    if let Some(key) = req.header.find_field("Sec-WebSocket-Key") {
        let mut hasher = crypto::sha1::Sha1::new();
        hasher.input_str(key);
        hasher.input_str("258EAFA5-E914-47DA-95CA-C5AB0DC85B11");

        let mut result = [0u8; 160 / 8];
        hasher.result(&mut result);
        let key = base64::encode(result);
        res.header.add_field("Sec-WebSocket-Accept", key.as_str());
    } else {
        return Err(Error::new(Kind::WebSocketError, Class::ClientError)
            .set_desc("unable to find header field 'Sec-WebSocket-Key'".to_string()));
    }

    client.server_keep_alive = false;
    res.header.add_field("Connection", "Upgrade");
    res.header.add_field("Upgrade", "websocket");

    res.status(101);
    res.send(&mut client.stream)?;

    Ok(())
}

pub fn error_handler(error: Error) {
    // TODO send error response frame
}

pub fn connection_handler(
    client: &mut http::HttpStream,
    req: &http::Request,
    mut res: http::Response,
) {
    if let Err(error) = handshake(client, req, &mut res) {
        return http::error_handler(client, res, error);
    }

    loop {
        match recv_message(client) {
            Ok(msg) => {
                match msg {
                    Message::TextMessage(msg) => {
                        // TODO threads?
                        let req: RequestEnvelope = serde_json::from_str(msg.data.as_str()).unwrap();
                        println!("Endpoint: {}, ReqNo: {}, Data: {}", req.endpoint, req.request_nr, req.data);
                        //let a = usimp::endpoint(req.endpoint.as_str(), req.data).unwrap();
                    },
                    Message::CloseMessage(msg) => {
                        println!("Received close frame: {}: {}", msg.code.unwrap_or(0), msg.reason.unwrap_or("-".to_string()));
                        return
                    }
                    Message::PingMessage(msg) => {
                        // TODO send pong
                    },
                    Message::PongMessage(msg) => {
                        // TODO something
                    }
                }
            },
            Err(error) => {
                error_handler(error);
            }
        }
    }
}
